package nplay.json.hornetq.client;

import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import nplay.common.app.Queue;

import org.hornetq.api.core.HornetQException;
import org.hornetq.api.core.SimpleString;
import org.hornetq.api.core.client.ClientConsumer;
import org.hornetq.api.core.client.ClientMessage;
import org.hornetq.api.core.client.ClientProducer;
import org.hornetq.api.core.client.ClientSession;
import org.hornetq.api.core.client.ClientSessionFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * @author Alfred Wong
 * 
 */
public class HornetQQueueSession {

	private static Logger logger = LoggerFactory.getLogger(HornetQQueueSession.class);
	private static Logger produceLogger = LoggerFactory.getLogger("produced");
	private static Logger consumedLogger = LoggerFactory.getLogger("consumed");
	private final String JMS_QUEUE_PREFIX = "jms.queue.";
	private final String MESSAGE_PARAM_NAME = "msg";
	
	private Queue queue;
	private SimpleString jmsQueueName;
	private ClientSessionFactory clientSessionFactory;
	private ConcurrentLinkedQueue<ClientSession> clientSessionConcurrentLinkQueue;
	private ClientSession clientSessionForConsumer;
	private ConcurrentLinkedQueue<ClientProducer> clientProducerConcurrentLinkQueue;
	private ClientConsumer clientConsumer;
	
	private int totalProducers ;
	private long repollProducerInterval;
	private int consumeTimeout;
	private AtomicBoolean isExceptionThrowed;
	private AtomicBoolean isClosed;
	
	// Default variables
	
	private final int DEFAULT_EXPIRATION = 0; // Count in ms
    private boolean autoCommitSends = true;
    private boolean autoCommitAcks = true;
	
    /*
     * Constructor with parameters
     * 
     * @param queue The queue which hornet queue session create for
     * @param clientSessionFactory HornetQ client session factory
     * @param totalProducers Number of producer to be create to support higher TPS in produce
     * @param repollProducerInterval Interval to poll again producer from the concurrent link queue
     * @param consumeTimeout Timeout for consume message from HornetQ server
     */
	public HornetQQueueSession(Queue queue, ClientSessionFactory clientSessionFactory, int totalProducers, int repollProducerInterval, int consumeTimeout) throws HornetQException {
		this.queue = queue;
		this.jmsQueueName= new SimpleString(JMS_QUEUE_PREFIX + queue.name()); 
		this.clientSessionFactory = clientSessionFactory;
		this.totalProducers = totalProducers;
		this.repollProducerInterval = repollProducerInterval;
		this.consumeTimeout = consumeTimeout;
		
		clientSessionConcurrentLinkQueue = new ConcurrentLinkedQueue<ClientSession>();
		clientProducerConcurrentLinkQueue = new ConcurrentLinkedQueue<ClientProducer>();
		isExceptionThrowed = new AtomicBoolean(false);
		isClosed = new AtomicBoolean(false);
		
		initQueueSession();
	}
	
	/*
	 * Method used to initialize variables
	 */
	private void initQueueSession() throws HornetQException {
		try {
			// Create producer(s) from client session(s)
			// Block on durable send guarantee message arrived at HornetQ server side,
			// but each producer only able to handle 100-150(max) TPS, to overcome that 
			// we can create more producer but each from separate session as required
			// by HornetQ, e.g. 2xProducers able to get TPS 200-300..
			for(int i=0; i<totalProducers; i++) {
				// Set autoCommit and autoAck on send
				ClientSession clientSessionForProducer = clientSessionFactory.createSession(autoCommitSends, autoCommitAcks);
				clientSessionConcurrentLinkQueue.add(clientSessionForProducer);
				ClientProducer clientProducer = clientSessionForProducer.createProducer(jmsQueueName);
				clientProducerConcurrentLinkQueue.add(clientProducer);
			}
			
			// Create consumer, unlike producer, single consumer can consume message in TPS 400-500
			// autoCommitAcks here doesn't meant auto acknowledge, it just mean it will commit the ack
			// after fire by the code, so after consume each message we still need to ack it MANUALY!
			clientSessionForConsumer = clientSessionFactory.createSession(autoCommitSends, autoCommitAcks);
			// Consumer session MUST be start to able consume messages from HornetQ server!
			clientSessionForConsumer.start();
			clientConsumer = clientSessionForConsumer.createConsumer(jmsQueueName);
		} catch (HornetQException e) {
			logger.error("Failed to initialization parameters for [Queue, HornetQErrorCode] {}", new Object[] { queue.name(), e.getCode() }, e);
			// Closed resources properly
			closeQueueSession();
			throw e;
		}
	}
	
	/*
	 * Method used to produce message with default expiration
	 * 
	 * @param queueMessage The JSON string message
	 * @param priority Message priority
	 */
	public void produceTextMessage(String queueMessage, int priority) throws HornetQException, HornetQClientException {
		produceTextMessage(queueMessage, priority, DEFAULT_EXPIRATION);
	}
	
	/*
	 * Method used to produce message
	 * 
	 * @param queueMessage The JSON string message
	 * @param priority Message priority
	 * @param expiration How long time a message last before change to expire
	 */
	public void produceTextMessage(String queueMessage, int priority, long expiration) throws HornetQException, HornetQClientException {
		// Check whether should stop process produce request
		checkIsClosed();
		// Create durable message, can peek either one client session (producer) from concurrent link queue
		ClientMessage clientMessage = clientSessionConcurrentLinkQueue.peek().createMessage(true);
		// Set message to be send in particular property in the client message
		clientMessage.putStringProperty(MESSAGE_PARAM_NAME, queueMessage);
		// Set the message priority, it must be cast to byte due to API restriction
		clientMessage.setPriority((byte)priority);
		// Set message expiration
		clientMessage.setExpiration(expiration);
		// WONG: TESTING
//		clientMessage.putStringProperty(Message.HDR_DUPLICATE_DETECTION_ID.toString(), UUID.randomUUID().toString());
		try {
			ClientProducer clientProducer = null;
			do {
				// To avoid infinite loop in situation where previous thread hit exception and not put back
				// client producer polled out back to queue
				if(isExceptionThrowed.get()) {
					throw new HornetQClientException("REJECT PRODUCE: Message[{}] Priority[{}] Expiration[{}] due to exception throwed recently..");
				}
				// Poll and remove client producer from the queue
				clientProducer = clientProducerConcurrentLinkQueue.poll();
				// This might happened when all client producers in the queue are currently used by other threads
				if(clientProducer == null) {
					try {
						// Wait for specific interval before retry poll client producer again
						Thread.sleep(repollProducerInterval);
					} catch (InterruptedException e) {
					}
				} else {
					// Break while loop if producer poll successfully from queue
					break;
				}
			} while (true);
			synchronized (clientProducer) {
				// Send message using client producer
				clientProducer.send(clientMessage);
				// Put back the client producer to the queue for usage by other threads
				clientProducerConcurrentLinkQueue.add(clientProducer);
			}
			// Log the successful produce details to a specific log
			produceLogger.info("PRODUCE: Queue[{}] Message[{}] Priority[{}] Expiration[{}]", new Object[] { queue.name(), queueMessage, priority, expiration } );
		} catch (HornetQException e) {
			// Set exception throw to true, please reference produce logic for the purpose to do this
			isExceptionThrowed.set(true);
			// Log the fail produce details to common log
			logger.error("PRODUCE FAILED: [Queue, Message, Priority, Expiration, HornetQErrorCode] {}", new Object[] { queue.name(), queueMessage, priority, expiration, e.getCode() }, e);
			throw e;
		}
	}
	
	/*
	 * Method used to consume message from HornetQ server
	 * 
	 * @return The string message take out from the consumed client message from HornetQ server
	 */
	public String consumeTextMessage() throws HornetQException {
		ClientMessage clientMessage;
		try {
			synchronized (clientConsumer) {
				// Must set a timeout or else sometime it will hang here because still waiting response from HornetQ
				clientMessage = clientConsumer.receive(consumeTimeout);
				if(clientMessage != null) {
					// Core API need to manually acknowledge the message or else message will be retain in the queue
					clientMessage.acknowledge();
					// Get the message in string mode from particular property in the consumed message
					String text = clientMessage.getStringProperty(MESSAGE_PARAM_NAME);
					// Log the consumed message details in a specific log
					consumedLogger.info("CONSUMED: Queue[{}] Message[{}] Priority[{}] Expiration[{}]", new Object[] { queue.name(), text, clientMessage.getPriority(), clientMessage.getExpiration() } );
					return text;
				}
			}
		} catch (HornetQException e) {
			logger.error("FAILED CONSUME FROM [Queue, HornetQErrorCode] {}", new Object[] { queue.name(), e.getCode() }, e);
			throw e;
		}
		return null;
	}
	
	/*
	 * Method used to close the queue session and it child object free up resources
	 */
	public synchronized void closeQueueSession() {
		// Set close to true to stop any produce or consume request
		isClosed.set(true);
		// Close client producer(s) to release HornetQ resources properly
		Iterator<ClientProducer> itClientProducer = clientProducerConcurrentLinkQueue.iterator();
		while(itClientProducer.hasNext()) {
			ClientProducer clientProducer = itClientProducer.next();
			if(clientProducer != null) {
				try {
					clientProducer.close();
				} catch (HornetQException e) {
					logger.error("Failed to close client producer on " + queue.name(), e);
				}
			}
		}
		// Close client session(s) for producer(s)
		Iterator<ClientSession> itClientSession = clientSessionConcurrentLinkQueue.iterator();
		while(itClientSession.hasNext()) {
			ClientSession clientSession = itClientSession.next();
			if(clientSession != null) {
				try {
					clientSession.close();
				} catch (HornetQException e) {
					logger.error("Failed to close client session (producer) on " + queue.name(), e);
				}
			}
		}
		// Close consumer
		if(clientConsumer != null) {
			try {
				clientConsumer.close();
			} catch (HornetQException e) {
				logger.error("Failed to close client consumer on " + queue.name(), e);
			}
		}
		// Close session used to create consumer
		if(clientSessionForConsumer != null) {
			try {
				clientSessionForConsumer.close();
			} catch (HornetQException e) {
				logger.error("Failed to close client session (consumer) on " + queue.name(), e);
			}
		}
		logger.info("HornetQQueueSession for '{}' closed", queue.name());
	}
	
	/*
	 * Common method used to check whether current Hornet queue session has been closed
	 */
	private void checkIsClosed() throws HornetQClientException {
		if(isClosed.get()) {
			// Let caller decide how to handle it
			throw new HornetQClientException("HornetQQueueSession has been closed...");
		}
	}
	
}
